"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var Cmd = require("../cmd");
var index_1 = require("../index");
tslib_1.__exportStar(require("./hash"), exports);
tslib_1.__exportStar(require("./memoize"), exports);
var utils_1 = require("../utils");
var dispatcher_1 = require("../dispatcher");
/**
 * ADT Helper for TS
 * e.g.
 * ```ts
 * type Msg =
 * | Dt<'fetchBook', number>
 * | Dt<'updateBook', Book>
 *
 * let msg = dt('fetchBook', 1)
 * switch(msg.tag) {
 *   case 'fetchBook':
 *      //...
 *      break
 *   case 'updateBook':
 *      //...
 *      break
 *   default:
 *      never(msg) // incomplete check from TS
 *      break
 * }
 * ```
 */
function dt(tag, data) {
    if (data === void 0) { data = null; }
    return { tag: tag, data: data };
}
exports.dt = dt;
/** @public */
exports.never = function (f) { return f; };
/** @public */
function mkInit(state, cmd) {
    if (cmd === void 0) { cmd = Cmd.none; }
    return { state: state, cmd: cmd };
}
exports.mkInit = mkInit;
function compose() {
    var fns = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        fns[_i] = arguments[_i];
    }
    return function (arg) { return fns.reduce(function (arg, fn) { return fn(arg); }, arg); };
}
exports.compose = compose;
/** @public */
function defaults(value, defaults) {
    return value == null ? defaults : value;
}
exports.defaults = defaults;
function combine(arg, _acts) {
    if (_acts === void 0) { _acts = null; }
    var state = {};
    var cmd = Cmd.none;
    var cmds = {};
    var actions = {};
    var views = {};
    var _loop_1 = function (key) {
        var comp = arg[key][0];
        var init = index_1.normalize(arg[key][1]);
        state[key] = init.state;
        actions[key] = comp.actions;
        views[key] = comp.view;
        cmds[key] = Cmd.map(function (_) { return _[key]; }, init.cmd);
        cmd = Cmd.batch(cmd, cmds[key]);
    };
    for (var key in arg) {
        _loop_1(key);
    }
    return {
        state: state,
        cmd: cmd,
        cmds: cmds,
        views: views,
        actions: actions,
        render: function (k, state, actions) {
            var view = views[k];
            if (view.length === 1) {
                return view({ state: state[k], actions: actions[k] });
            }
            else {
                return view(state[k], actions[k]);
            }
        },
    };
}
exports.combine = combine;
/**
 * @internal
 * run action and return a normalized result ([State, CmdType<>]),
 * this is useful to write High-Order-Action, which take an action and return a wrapped action.
 * @param result result of `action(msg: Data)`
 * @param state
 * @param actions
 */
function runActionResult(result) {
    var res = result;
    var _a = dispatcher_1.dispatcher.getContext(), state = _a.state, actions = _a.actions, parentState = _a.parentState, parentActions = _a.parentActions;
    utils_1.isFn(res) &&
        (res = res.call(actions, state, actions, parentState, parentActions)) &&
        utils_1.isFn(res) &&
        (res = res.call(actions, actions));
    // action can be a function that return a promise or undefined(callback)
    return index_1.normalize(res, state);
}
exports.runActionResult = runActionResult;
/**
 * Wrap a child action with parentState, parentActions.
 * @internal
 * @deprecated Deprecated for `overrideAction`
 * @param action The action to be wrapped
 * @param wrapper
 * @param parentState
 * @param parentActions
 */
function withParents(action, wrapper) {
    if (!wrapper) {
        return action;
    }
    var wrapped = function (state, actions, parentState, parentActions) {
        var nactions = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return runActionResult(action.apply(void 0, args));
        };
        return wrapper(nactions, parentState, parentActions, state, actions);
    };
    return wrapped;
}
exports.withParents = withParents;
/**
 * @deprecated Deprecated for `overrideAction`
 * @internal
 */
exports.wrapActions = withParents;
/**
 * Wrap a child action with parentState, parentActions.
 * @internal
 * @param action The action to be wrapped
 * @param wrapper
 * @param parentState
 * @param parentActions
 */
function overrideAction(parentActions, getter, wrapper) {
    if (!wrapper) {
        return parentActions;
    }
    var action = getter(parentActions);
    var wrapped = function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        var normalAction = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            var ret = runActionResult(action.apply(void 0, args));
            return ret;
        };
        var ctx = dispatcher_1.dispatcher.getContext();
        return wrapper.apply(void 0, args)(normalAction, ctx.parentState, ctx.parentActions, ctx.state, ctx.actions);
    };
    var keys = (getter.toString().match(/((?:[\w_$]+\.)+[\w_$]+)/) || [])[1].split('.').slice(1);
    var cursor = parentActions;
    var replaced = false;
    for (var i = 0; i < keys.length; i++) {
        var key = keys[i];
        if (cursor[key] === action) {
            cursor[key] = wrapped;
            replaced = true;
            break;
        }
        cursor = cursor[key] = tslib_1.__assign({}, cursor[key]);
    }
    utils_1.weakVal(wrapped, utils_1.OverrideLength, keys.length);
    if (!replaced) {
        console.error(new Error("Cannot find action in parentActions"), parentActions, getter);
    }
    return parentActions;
}
exports.overrideAction = overrideAction;
//# sourceMappingURL=index.js.map